#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>

#include "config.h"
#include "pidfile.hpp"
#include "server.hpp"
#include "conf.hpp"

#define DEFAULT_SERVERPORT 1099
#define DEFAULT_MAXPROCS	  10

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
// union semun is defined by including <sys/sem.h>
#else
// according to X/OPEN we have to define it ourselves
union semun {
	int val;                    // value for SETVAL
	struct semid_ds *buf;       // buffer for IPC_STAT, IPC_SET
	unsigned short int *array;  // array for GETALL, SETALL
	struct seminfo *__buf;      // buffer for IPC_INFO
};
#endif

// Function prototypes
bool AllowHost(struct in_addr);
void SignalProc(int);
void dprintf(char *, ...);
char *naddr2str(struct sockaddr_in *);
unsigned long int nname2h(char *, bool *);

// Global variables
struct sockaddr_in serv;
struct sockaddr_in child;
char szConfigFile[256] = "/etc/passwdd.conf";
char *pszConfig = szConfigFile;

int iServerPort;
int iMaxProcesses;

int listenfd, newlistenfd;
int sockfd, counterfd;

key_t idCounter = 1099;

CConfigurationFile *pConfigFile = NULL;
CPidFile *pPidFile = NULL;

int main(int argc, char **argv)
{
	int pidstatus;
	int child_process;
//	int err;
	char *pszResult;
	
	if (argc > 1)
		for (int i = 1; i < argc; i++)
			if ('-' == argv[i][0])				// We have option
				if ('c' == argv[i][1])			// We have other config file
					if ('\0' != argv[i][2]) {	// The user has not typed space after -c
						pszConfig = &argv[i][3];
						argc -= 1;
					} else {
						pszConfig = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
						argc -= 2;
					}
	
	if ((argc != 1) || (NULL == pszConfig)) {
		printf("Usage: passwdd [-c <config file>]\n");
		printf("Options: -c <config file> - Use alternate config file instead\n");
		printf("  the default '/etc/passwdd.conf' system wide config file.\n");
		printf("Written by: <Alexander Feldman> alex@varna.net\n\n");
	}
	
	pPidFile = new CPidFile("passwdd.pid");
	if (NULL == pPidFile) {
		fprintf(stderr, "passwdd: error allocating memory.\n");
		return 1;
	}

	dprintf("Checking pidfile.\n");
	if (true == pPidFile->CheckPid()) {
#ifndef DEBUG
		pid_t pid;
		if ((pid = fork()) > 0)
			exit(0);								// Parent process
		else
			if (pid < 0) {						// An error occured in the fork
				fprintf(stderr, "passwdd: initial fork error:%s.\n", strerror(errno));
				exit(1);
			}
		int i = getdtablesize();
		for (int j = 0; j < i; j++)
		  close(j);
		setsid();								// Creates a session and closes the controlling tty
#endif
	} else {
		fputs("passwdd: already running.\n", stderr);
		return 1;
	}
	dprintf("Writing pidfile.\n");
	if (true == pPidFile->CheckPid()) {
		if (false == pPidFile->WritePid()) {
			dprintf("Can't write pid file.\n");
			return 1;
		}
	} else {
		dprintf("Pidfile (and pid) already exists.\n");
		return 1;
	}
	
	signal(SIGTERM, SignalProc);
	signal(SIGHUP, SignalProc);
	signal(SIGCHLD, SignalProc);
	signal(SIGINT, SignalProc);
	signal(SIGQUIT, SignalProc);
	dprintf("Starting.\n");
	
	openlog(argv[0], LOG_PID, LOG_USER);
	syslog(LOG_USER | LOG_INFO, "passwdd daemon starting");
	
	pConfigFile = new CConfigurationFile(pszConfig, &pszResult);
	if (NULL == pConfigFile) {
		syslog(LOG_USER | LOG_INFO, "error allocating memory");
		return 1;
	}
	if (NULL != pszResult) {
		syslog(LOG_USER | LOG_INFO, "conf error: %s", pszResult);
		return 1;
	}
	
	if ((listenfd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		syslog(LOG_USER | LOG_INFO, "socket error: %s", strerror(errno));
		exit(1);
	}
	
	iServerPort = pConfigFile->GetInteger("Global", "port", DEFAULT_SERVERPORT);
	iMaxProcesses = pConfigFile->GetInteger("Global", "max_processes", DEFAULT_MAXPROCS);
	
	memset(&serv, 0, sizeof(serv));
	serv.sin_family = AF_INET;
	serv.sin_addr.s_addr = htonl(INADDR_ANY);
	serv.sin_port = htons(iServerPort);
	
	if (bind(listenfd, (struct sockaddr *)&serv, sizeof(serv)) < 0) {
		syslog(LOG_USER | LOG_INFO, "bind error: %s", strerror(errno));
		exit(1);
	}
	
	if (listen(listenfd, SOMAXCONN) < 0) {
		syslog(LOG_USER | LOG_INFO, "listen error: %s", strerror(errno));
		exit(1);
	}
	
	union semun arg;	
	while (-1 == (counterfd = semget(idCounter, 1, IPC_CREAT | IPC_EXCL | 0660))) {
		syslog(LOG_USER | LOG_INFO, "semget error for key %ld: %s", idCounter, strerror(errno));
		idCounter += 1;
		if (idCounter - 1099 > 32) {
			syslog(LOG_USER | LOG_INFO, "unable to create new semaphore");
			exit(1);
		}
	}
	arg.val = iMaxProcesses;
	if (-1 == semctl(counterfd, 0, SETVAL, arg)) {
		syslog(LOG_USER | LOG_INFO, "semctl error: %s", strerror(errno));
		exit(1);
	}
	
	child_process = 0;
	while (0 == child_process) {			// Loop begins here
		memset(&child, 0, sizeof(child));
		int childlen = (int)sizeof(child);
		if ((sockfd = accept(listenfd, (struct sockaddr *)&child, &childlen)) < 0) {
			if (EINTR != errno)
				syslog(LOG_USER | LOG_INFO, "accept error (%s): %s", naddr2str(&child), strerror(errno));
			continue;
		}
		syslog(LOG_USER | LOG_INFO, "opened client connection from %s", naddr2str(&child));
		if (false == AllowHost(child.sin_addr)) {
			syslog(LOG_USER | LOG_INFO, "unauthorized client connection from %s", naddr2str(&child));
			close(sockfd);
			continue;
		}
		pid_t pid;
		if ((pid = fork()) == 0) {			// This is the child
			if (-1 == (counterfd = semget(idCounter, 1, 0))) {
				syslog(LOG_USER | LOG_INFO, "semget error: %s", strerror(errno));
				close(listenfd);
				close(sockfd);
				return 0;
			}
			struct sembuf sbuf = { 0, -1, SEM_UNDO };			
			if (-1 == semop(counterfd, &sbuf, 1)) {
				syslog(LOG_USER | LOG_INFO, "semget error: %s", strerror(errno));
				close(listenfd);
				close(sockfd);
				return 0;
			}
			
			close(listenfd);
			server(sockfd);
			close(sockfd);
			child_process = 1;				// Exit from main loop for the child process
		} else {									// We are parent, go on listening and error checking
			if (pid < 0)
				syslog(LOG_USER | LOG_INFO, "fork error: %s", strerror(errno));
			while(waitpid(-1, &pidstatus, WNOHANG) > 0);
			close(sockfd);
			child_process = 0;
		}
	}
	return 0;
}

bool AllowHost(struct in_addr lSource)
{
	register int i, j;
	static char szBuf[256];
	static char szTmp[256];
	char *pszHost;
	char *pszMask;
// We will compare the source address, the target address and target netmask
// all in host byte order. There is no matter what ordering we will use
// but is logically more correct to perform it in native ordering.	
	unsigned long int ulSourceHost = ntohl(lSource.s_addr);
	unsigned long int ulTargetHost;
	unsigned long int ulTargetMask;
	register unsigned long l, q;
	char *p;
	bool fgError;
	
	char *pszAllowedHosts = pConfigFile->GetString("Permissions", "hosts_allow", "localhost");
	for (i = j = 0; pszAllowedHosts[i] != '\0'; i++)
	   if (NULL == strchr(" \t\x0a\x0d", pszAllowedHosts[i]))
		   szBuf[j++] = pszAllowedHosts[i];
	szBuf[j] = '\0';

	if (NULL != (pszHost = strtok(szBuf, ",;")))
		do {
			pszHost = strcpy(szTmp, pszHost);
			if (NULL != (pszMask = strchr(pszHost, '/'))) {
				*pszMask = '\0';
				pszMask += 1;
			}
			ulTargetHost = nname2h(*pszHost == '!' ? pszHost + 1 : pszHost, &fgError);
			if (true == fgError) {
				syslog(LOG_USER | LOG_INFO, "passwdd: configuration file error, skipping host");
				continue;
			}
			ulTargetMask = 0xFFFFFFFF;
			if (NULL != pszMask) {
				l = strtol(pszMask, &p, 10);
				if (('\0' == *p) && (l <= 32)) {
					ulTargetMask = 0;
					for (q = 0; q < l; q++) {
						ulTargetMask >>= 1;
						ulTargetMask |= 0x80000000;
					}
				} else {
					ulTargetMask = nname2h(pszMask, &fgError);
					if (true == fgError) {
						syslog(LOG_USER | LOG_INFO, "passwdd: configuration file error, skipping network");
						continue;
					}
			   }
			}
			if ('!' == *pszHost)
				if ((ulSourceHost & ulTargetMask) == ulTargetHost)
					return false;
			if ((ulSourceHost & ulTargetMask) == ulTargetHost)
				return true;
		} while (NULL != (pszHost = strtok(NULL, ",;")));
	
	return false;
}

void SignalProc(int iSignal)
{
	CConfigurationFile *pNewConfigFile;	
	char *pszResult;
	int iPidStatus;	
	
	switch (iSignal) {
		case SIGCHLD:
			while(waitpid(-1, &iPidStatus, WNOHANG) > 0);
		break;
		case SIGHUP:
			syslog(LOG_USER | LOG_INFO, "reinitializing");		
			pNewConfigFile = new CConfigurationFile(pszConfig, &pszResult);
			if (NULL == pNewConfigFile) {
				syslog(LOG_USER | LOG_INFO, "error allocating memory");
				return;
			}
			if (NULL != pszResult) {
				syslog(LOG_USER | LOG_INFO, "conf error: %s", pszResult);
				delete pNewConfigFile;
				return;
			}
			pConfigFile = pNewConfigFile;
			if (iServerPort != pConfigFile->GetInteger("Global", "port", DEFAULT_SERVERPORT)) {
				iServerPort = pConfigFile->GetInteger("Global", "port", DEFAULT_SERVERPORT);
				if ((newlistenfd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
					syslog(LOG_USER | LOG_INFO, "socket error: %s", strerror(errno));
					return;
				}
				memset(&serv, 0, sizeof(serv));
				serv.sin_family = AF_INET;
				serv.sin_addr.s_addr = htonl(INADDR_ANY);
				serv.sin_port = htons(iServerPort);
	
				if (bind(newlistenfd, (struct sockaddr *)&serv, sizeof(serv)) < 0) {
					syslog(LOG_USER | LOG_INFO, "bind error: %s", strerror(errno));
					return;
				}
	
				if (listen(newlistenfd, SOMAXCONN) < 0) {
					syslog(LOG_USER | LOG_INFO, "listen error: %s", strerror(errno));
					return;
				}
				close(listenfd);
				listenfd = newlistenfd;
			}
		break;
		case SIGTERM:
		case SIGINT:
		case SIGQUIT:
			union semun arg;
			if (-1 != (counterfd = semget(idCounter, 1, 0)))
				semctl(counterfd, 0, IPC_RMID, arg);
		
			delete pConfigFile;
			delete pPidFile;
			if (0 != iSignal)
				dprintf("exiting on signal %d\n", iSignal);
		
			exit(0);
		break;
	}
	signal(iSignal, SignalProc);			// Reset to get it again	
}

void dprintf(char *
#ifdef DEBUG
				 fmt
#endif
				 , ...)
{
#ifdef DEBUG	
	va_list ap;
	
	va_start(ap, fmt);
	vfprintf(stdout, fmt, ap);
	va_end(ap);
	
	fflush(stdout);
#endif
}

char *naddr2str(struct sockaddr_in *saptr) {
	static char szResult[256];
	char	*bp, *ap;
	int l;
	struct sockaddr sa;
	
// Check for null/zero family
	if (saptr == NULL)
		return "NULLADDR";
	if (saptr->sin_family == 0)
		return "0";

	switch (saptr->sin_family) {
//		case AF_UNIX:
//			if (saptr->sunix.sun_path[0] != '\0')
//				snprintf(str, 256, "[UNIX: %.64s]", saptr->sunix.sun_path);
//			else
//				snprintf(str, 256, "[UNIX: localhost]");
//			return str;
		case AF_INET:
			return inet_ntoa(saptr->sin_addr);
	}

// Unknown family -- just dump bytes
	memcpy(&sa, saptr, sizeof(sa));
	snprintf(szResult, 256, "Family %d: ", sa.sa_family);
   bp = &szResult[strlen(szResult)];
   ap = sa.sa_data;
   for (l = sizeof(sa.sa_data); --l >= 0; ) {
		sprintf(bp, "%02x:", *ap++ & 0377);
		bp += 3;
	}
	*--bp = '\0';
	return szResult;
}

unsigned long nname2h(char *pszName, bool *pfgError)
{
	struct hostent *pstrHost;
	
	pstrHost = gethostbyname(pszName);
	if (NULL == pstrHost) {
		*pfgError = true;
		return 0;
	}
	*pfgError = false;
	return ntohl(*((unsigned long int *)(pstrHost->h_addr)));
}